;+
; NAME:
;     ncregrid
;
; PURPOSE:
;     Interpolate or regrid all data fields of one netcdf file from
;     one rectangular grid to another grid. The default method for
;     regridding is bilinear interpolation. For emission fluxes and
;     other quantities that need to be mass conserved, use the
;     conserve_mass keyword. 
;
; CALLING SEQUENCE:
;     ncregrid, filename, oldgridname, newgridname,  $
;            oshiftlon=oshiftlon, nshiftlon=nshiftlon,  $
;            conserve_mass=conserve_mass, add_area=add_area,  $
;            outfile=outfile, quiet=quiet
;
; ARGUMENTS:
;     filename (input, string) -> the name of the netcdf file to be
;          regridded. 
;     oldgridname (input, string) -> the name of the original (input)
;          grid. This name must be contained in the mgs_rgrid
;          catalogue. See mgs_rgrid__define.pro for definition of
;          additional grids. You can use the oshiftlon keyword to change
;          the starting longitude of the default definition.
;     newgridname (input, string) -> the name of the output grid. This
;          grid must be defined in the mgs_rgrid catalogue. You can
;          use the nshiftlon keyword to change the starting longitude of
;          the default definition. 
;     oshiftlon (input, float) -> offset for shifting the starting
;          longitude of the input grid. Most grids are defined
;          starting at 0 degree longitude. Set oshiftlon=-180. to
;          obtain a grid starting at 180 degrees west.
;     nshiftlon (input, float) -> offset for shifting the starting
;          longitude of the output grid.
;     oreverse_latitude (input, boolean) -> reverse latitudes of input
;          grid (normally, most grids are defined south to north).
;     nreverse_latitude (input, boolean) -> reverse latitudes of
;          output grid.
;     conserve_mass (input, boolean) -> use a slower interpolation
;          method which computed the new grid box value as weighted
;          average over all underlying grid boxes. Contrary to the
;          bilinear interpolation, this method is strictly mass
;          conserving. 
;     add_area (input, boolean) -> add a variable named gridbox_area
;          if not already present.
;     outfile (input, string) -> name of the output file. If not
;          given, ncregrid will use the input filename and replace
;          ".nc" with ".[newgridname].nc"
;     quiet (input, boolean) -> Set this keyword to suppress the
;          progress messages.
;
; LIMITATIONS:
;     The fields that can be regridded can only have up to four
;     dimensions including the longitude, latitude, and a time
;     dimension. Longitude, latitude, and time are identified by their
;     name (a few variants are possible), and one extra dimension is
;     allowed (typically some sort of level index). The file must be
;     small enough to fit into memory in both the original and the
;     output resolution. 
;     Note that the attributes for longitude and latitude will be
;     overwritten. 
;     If 4D variables are to be regridded, their 4th dimension should
;     not be lon or lat.
;
; MODIFICATION HISTORY:
;     Martin Schultz, MPI Meteorology, Hamburg, 14 Feb 2005: VERSION 1.0
;     mgs, 15 Feb 2005 : - include global attributes
;     mgs, 10 Mar 2005 : - bug fix in gridbox_area handling. Now makes
;                          sure that gridbix_area variable is not regridded.
;-

pro ncregrid, filename, oldgridname, newgridname, oshiftlon=oshiftlon, $
              nshiftlon=nshiftlon, oreverse_latitude=oreverse_latitude, $
              nreverse_latitude=nreverse_latitude, conserve_mass=conserve_mass, $
              add_area=add_area, outfile=outfile, quiet=quiet

   quiet = keyword_set(quiet)

   ;; check validity of old and new grid names
   oldgrid = obj_new('mgs_rgrid', gridname=oldgridname, shiftlon=oshiftlon, $
                    reverse_lat=oreverse_latitude)
   if obj_valid(oldgrid) eq 0 then return
   newgrid = obj_new('mgs_rgrid', gridname=newgridname, shiftlon=nshiftlon, $
                    reverse_lat=nreverse_latitude)
   if obj_valid(newgrid) eq 0 then return

   ;; set regridding keywords
   interpolate = 1
   area_weighted = 0
   if keyword_set(conserve_mass) then begin
      interpolate = 0
      area_weighted = 1
   endif 

   ;; read the netcdf file
   ncdf_read, odata, file=filename, /all, /aspointer, attr=attr, $
      dimnames=dimnames, vardimid=vardimid, truename=truename
   vn = tag_names(odata)
   vdimn = tag_names(vardimid)
   an = tag_names(attr)

   ;; create output structures
   ndata = { lon:newgrid->GetLon(), $
             lat:newgrid->GetLat() }
   lonattr = { long_name:'longitude', units:'degrees_east' }
   latattr = { long_name:'latitude', units:'degrees_north' }
   nattr = { lon:lonattr, $
             lat:latattr }

   ;; add global attributes
   if chkstru(attr, 'global') then begin
      nattr = create_struct(nattr, 'global', attr.global)
   endif 

   ;; find out which dimension is lon and lat
   wlon = where(strlowcase(dimnames) eq 'lon' or $
                strlowcase(dimnames) eq 'longitude' or $
                strlowcase(dimnames) eq 'x', cnt)
   if cnt ne 1 then begin
      print, 'Cannot find longitude dimension in file!'
      return
   endif 
   wlon = wlon[0]
   wlat = where(strlowcase(dimnames) eq 'lat' or $
                strlowcase(dimnames) eq 'latitude' or $
                strlowcase(dimnames) eq 'y', cnt)
   if cnt ne 1 then begin
      print, 'Cannot find latitude dimension in file!'
      return
   endif 
   wlat = wlat[0]
   ;; check for record dimension (time)
   wtime = where(strlowcase(dimnames) eq 'time', cnt)
   if cnt gt 0 then unlimited = dimnames[wtime[0]]
   ;; note: unlimited remains undefined otherwise

   ;; loop over variables:
   ;; if lon and lat are used, then interpolate. If neither is used,
   ;; copy variable as is. If only one is used, produce an error
   ;; message. Ignore the lon and lat variables themselves.
   for i=0L, n_elements(vn)-1 do begin
      ;; find attributes for this variable
      wattr = where(an eq vn[i], cnta)
      ;; find dimensions of this variable
      wdim = where(vdimn eq vn[i], cnt)
      if cnt eq 0 then stop,'Strange error: vardimid not defined for '+vn[i]
      thisdims = vardimid.(wdim[0])
      haslon = (where(thisdims eq wlon))[0] ge 0
      haslat = (where(thisdims eq wlat))[0] ge 0
      if haslon eq 0 and haslat eq 0 then begin
         ;; variable has neither lon or lat dimension
         if not quiet then print, 'Copying variable '+vn[i]+' ...'
         if cnta gt 0 then nattr = create_struct(nattr, vn[i], attr.(wattr[0]))
         ndata = create_struct(ndata, vn[i], odata.(i))
      endif else if haslon gt 0 and haslat gt 0 then begin
         ;; variable is at least 2D with lon and lat
         ;; extract variable and regrid - loop over 4th dimension if present
         var = *odata.(i)
         vardims = size(var, /dimensions)
         ndims = n_elements(vardims)
         ;; skip gridbox_area variable 
         if strlowcase(vn[i]) eq 'gridbox_area' then begin
            print,'Skipping variable gridbox_area ...'
            continue
         endif 

         if ndims le 3 then begin
            ;; call regrid routine
            if interpolate then begin
               if not quiet then print, 'Interpolating '+vn[i]+' ...'
            endif else begin
               if not quiet then print,'Regridding '+vn[i]+' ...'
            endelse 
            resfield = regrid(var,oldgrid=oldgrid, newgrid=newgrid, /use_grids, $
                              interpolate=interpolate, area_weighted=area_weighted)
            ;; add to structure
            if cnta gt 0 then nattr = create_struct(nattr, vn[i], attr.(wattr[0]))
            ndata = create_struct(ndata, vn[i], ptr_new(resfield))
         endif else if ndims gt 4 then begin
            print, 'Sorry! Cannot regrid fields with dims>4!'
            print, 'Variable '+vn[i]+' will not be included in output.'
         endif else begin
            if interpolate then begin
               if not quiet then print, 'Interpolating 4D variable '+vn[i]+' ...'
            endif else begin
               if not quiet then print,'Regridding 4D variable '+vn[i]+' ...'
            endelse 
            ;; regrid first record to obtain resulting dimensions
            resfield = regrid(var[*,*,*,0],oldgrid=oldgrid, newgrid=newgrid, /use_grids, $
                              interpolate=interpolate, area_weighted=area_weighted)
            resdims = size(resfield, /dimensions)
            resfield = transpose(reform(resfield, resdims[0], resdims[1], resdims[2], 1))
            for j=1L,vardims[3]-1 do begin
               thisfield = regrid(var[*,*,*,j],oldgrid=oldgrid, newgrid=newgrid, /use_grids, $
                                 interpolate=interpolate, area_weighted=area_weighted)
               thisfield = reform(thisfield, resdims[0], resdims[1], resdims[2], 1)
               resfield = [ resfield, transpose(thisfield) ]
            end 
            ;; add to structure
            if cnta gt 0 then nattr = create_struct(nattr, vn[i], attr.(wattr[0]))
            ndata = create_struct(ndata, vn[i], ptr_new(transpose(resfield)))
         endelse 
      endif else begin
         ;; only lon or lat: ignore
         if strlowcase(vn[i]) eq 'lon' or $
            strlowcase(vn[i]) eq 'longitude' or $
            strlowcase(vn[i]) eq 'x' or $
            strlowcase(vn[i]) eq 'lat' or $
            strlowcase(vn[i]) eq 'latitude' or $
            strlowcase(vn[i]) eq 'y' then ok=1 else begin
            print,'Cannot transfer variable '+vn[i]+ $
               ' because it does have only one lon or lat dimension.'
         endelse 
      endelse 
   endfor 

   ;; add gridbox area if requested
   if keyword_set(add_area) then begin
      area = newgrid->GetArea()
      aattr = { long_name:'gridbox area', $
                units:'m^2' }
      nattr = create_struct(nattr, 'gridbox_area', aattr)
      ndata = create_struct(ndata, 'gridbox_area', ptr_new(area))
   endif 

   ;; write output file
   if n_elements(outfile) eq 0 then begin
      outfile = file_dirname(truename, /mark_directory) + $
         file_basename(truename,'.nc')+'.'+newgridname+'.nc'
   endif 
   history = systime(0)+': IDL ncregrid from '+oldgridname+ $
      ' to '+newgridname+' '+truename
   if not quiet then print, 'Writing interpolated data to '+outfile+' ...'
   ncdf_write, outfile, ndata, dimnames, unlimited=unlimited, $
      attr=nattr, history_entry=history
   if not quiet then print, 'Done.'

   ;; quick and dirty garbage cleaning
   heap_gc

end    

